/*
 * Gksu -- a library providing access to su functionality
 * Copyright (C) 2004 Gustavo Noronha Silva
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <pty.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <errno.h>

#include <glib.h>
#include <locale.h>

#include "defines.h"
#include "../config.h"

#include "gksu-context.h"

/* local function declarations */
static void
gksu_context_init (GTypeInstance *instance, gpointer g_class);

static void
gksu_context_dispose (GObject *object);

static void
gksu_context_finalize (GObject *object);

static void
gksu_context_class_init (GksuContextClass *klass);

void
nullify_password (GksuContext *context)
{
  if (context->password)
    g_free (context->password);
  context->password = NULL;
}

/**
 * gksu_context_new:
 *
 * Creates a new context to use gksu with. You must
 * manually free the context with gksu_context_free
 * after use.
 *
 * Returns: a new #GksuContext.
 */
GksuContext*
gksu_context_new ()
{
  g_type_init ();
  return (GksuContext*) g_type_create_instance (GKSU_TYPE_CONTEXT);
}

/**
 * gksu_context_free:
 * @context: the #GksuContext to be freed.
 *
 * Frees the provided #GksuContext.
 */
void
gksu_context_free (GksuContext *context)
{
  g_object_unref (G_OBJECT(context));
}

/**
 * gksu_context_set_user:
 * @context: the #GksuContext you want to modify
 * @username: the target username
 *
 * Sets up what user the command will be run as. The default
 * is root, but you can run the command as any user.
 *
 */
void
gksu_context_set_user (GksuContext *context, gchar *username)
{
  g_assert (username != NULL);

  if (context->user)
    g_free (context->user);
  context->user = g_strdup (username);
}

/**
 * gksu_context_get_user:
 * @context: the #GksuContext from which to grab the information
 *
 * Gets the user the command will be run as, as set
 * by gksu_context_set_user.
 *
 * Returns: a pointer to the string containing the username.
 */
const gchar*
gksu_context_get_user (GksuContext *context)
{
  return context->user;
}

/**
 * gksu_context_set_command:
 * @context: the #GksuContext you want to modify
 * @command: the command that shall be ran
 *
 * Sets up what command will run with the target user.
 *
 */
void
gksu_context_set_command (GksuContext *context, gchar *command)
{
  g_assert (command != NULL);

  if (context->command)
    g_free (context->command);
  context->command = g_strdup (command);
}

/**
 * gksu_context_get_command:
 * @context: the #GksuContext from which to grab the information
 *
 * Gets the command that will be run, as set by
 * gksu_context_set_command.
 *
 * Returns: a pointer to the string containing the command.
 */
const gchar*
gksu_context_get_command (GksuContext *context)
{
  return context->command;
}

/**
 * gksu_context_set_password:
 * @context: the #GksuContext you want to modify
 * @password: the password
 *
 * Sets up what is the password to be passed to su/sudo. *
 */
void
gksu_context_set_password (GksuContext *context, gchar *password)
{
  if (context->password)
    {
      g_free (context->password);
      context->password = NULL;
    }

  if (password && strcmp ("", password) && strcmp ("\n", password))
    {
      if (password[strlen(password) - 1] == '\n')
	context->password = g_strdup (password);
      else
	context->password = g_strdup_printf ("%s\n", password);
    }
}

/**
 * gksu_context_get_password:
 * @context: the #GksuContext from which to grab the information
 *
 * Gets the password that is set for calling the su/sudo
 * programs.
 *
 * Returns: a pointer to the string containing the password.
 */
const gchar*
gksu_context_get_password (GksuContext *context)
{
  return context->password;
}

/**
 * gksu_context_set_login_shell:
 * @context: the #GksuContext you want to modify
 * @value: TRUE or FALSE
 *
 * Should the shell in which the command will be run be
 * a login shell?
 */
void
gksu_context_set_login_shell (GksuContext *context, gboolean value)
{
  context->login_shell = value;
}

/**
 * gksu_context_get_login_shell:
 * @context: the #GksuContext from which to grab the information
 *
 * Finds out if the shell created by the underlying su process
 * will be a login shell.
 *
 * Returns: TRUE if the shell will be a login shell, FALSE otherwise.
 */
gboolean
gksu_context_get_login_shell (GksuContext *context)
{
  return context->login_shell;
}

/**
 * gksu_context_set_keep_env:
 * @context: the #GksuContext you want to modify
 * @value: TRUE or FALSE
 *
 * Should the environment be kept as it is? Defaults to
 * TRUE. Notice that setting this to FALSE may cause the
 * X authorization stuff to fail.
 */
void
gksu_context_set_keep_env (GksuContext *context, gboolean value)
{
  context->keep_env = value;
}

/**
 * gksu_context_get_keep_env:
 * @context: the #GksuContext from which to grab the information
 *
 * Finds out if the environment in which the program will be
 * run will be reset.
 *
 * Returns: TRUE if the environment is going to be kept,
 * FALSE otherwise.
 */
gboolean
gksu_context_get_keep_env (GksuContext *context)
{
  return context->keep_env;
}

/**
 * gksu_context_set_debug:
 * @context: the #GksuContext you want to modify
 * @value: TRUE or FALSE
 *
 * Set up if debuging information should be printed.
 */
void
gksu_context_set_debug (GksuContext *context, gboolean value)
{
  context->debug = value;
}

/**
 * gksu_context_get_debug:
 * @context: the #GksuContext from which to grab the information
 *
 * Finds out if the library is configured to print debuging
 * information.
 *
 * Returns: TRUE if it is, FALSE otherwise.
 */
gboolean
gksu_context_get_debug (GksuContext *context)
{
  return context->debug;
}

/**
 * gksu_context_set_ssh_fwd:
 * @context: the #GksuContext you want to modify
 * @value: TRUE or FALSE
 *
 * Set up if we are running inside a sshd forwarding
 * session. That's because the X authorization semantics
 * differ a bit.
 * -> THIS SETTING IS CURRENTLY USED FOR NOTHING
 */
void
gksu_context_set_ssh_fwd (GksuContext *context, gboolean value)
{
  context->ssh_fwd = value;
}

/**
 * gksu_context_get_ssh_fwd:
 * @context: the #GksuContext from which to grab the information
 *
 * Finds out if the library is configured as being run inside a
 * ssh X11 forwarding tunnel.
 * -> THIS SETTING IS CURRENTLY USED FOR NOTHING
 *
 * Returns: TRUE if it is, FALSE otherwise.
 */
gboolean
gksu_context_get_ssh_fwd (GksuContext *context)
{
  return context->ssh_fwd;
}

GType
gksu_context_get_type (void)
{
  static GType type = 0;

  if (type == 0)
    {
      static const GTypeInfo info =
	{
	  sizeof (GksuContextClass), /* size of class */
	  NULL, /* base_init */
	  NULL, /* base_finalize */
	  (GClassInitFunc) gksu_context_class_init,
	  NULL, /* class_finalize */
	  NULL, /* class_data */
	  sizeof (GksuContext), /* size of object */
	  0, /* n_preallocs */
	  (GInstanceInitFunc) gksu_context_init /* instance_init */
	};
      type = g_type_register_static (G_TYPE_OBJECT,
				     "GksuContextType",
				     &info, 0);
    }

  return type;
}

static void
gksu_context_init (GTypeInstance *instance, gpointer g_class)
{
  GksuContext *self = (GksuContext*)instance;

  self->xauth = NULL;
  self->dir = NULL;
  self->display = NULL;
  self->user = g_strdup ("root");
  self->password = NULL;
  self->command = NULL;

  self->login_shell = FALSE;
  self->keep_env = FALSE;

  self->debug = FALSE;
  self->ssh_fwd = FALSE;
}

static void
gksu_context_dispose (GObject *object)
{
  GksuContext *self = (GksuContext*)object;

  g_free (self->xauth);
  g_free (self->dir);
  g_free (self->display);
  g_free (self->user);
  g_free (self->password);
  g_free (self->command);
}

static void
gksu_context_finalize (GObject *object)
{
}

static void
gksu_context_class_init (GksuContextClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  gobject_class->dispose = gksu_context_dispose;
  gobject_class->finalize = gksu_context_finalize;

}

GType
gksu_context_error_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { GKSU_CONTEXT_ERROR_HELPER, "GKSU_CONTEXT_ERROR_HELPER", "helper" },
      { GKSU_CONTEXT_ERROR_NOCOMMAND, "GKSU_CONTEXT_ERROR_NOCOMMAND", "nocommand" },
      { GKSU_CONTEXT_ERROR_NOPASSWORD, "GKSU_CONTEXT_ERROR_NOPASSWORD", "nopassword" },
      { GKSU_CONTEXT_ERROR_FORK, "GKSU_CONTEXT_ERROR_FORK", "fork" },
      { GKSU_CONTEXT_ERROR_EXEC, "GKSU_CONTEXT_ERROR_EXEC", "exec" },
      { GKSU_CONTEXT_ERROR_PIPE, "GKSU_CONTEXT_ERROR_PIPE", "pipe" },
      { GKSU_CONTEXT_ERROR_PIPEREAD, "GKSU_CONTEXT_ERROR_PIPEREAD", "piperead" },
      { GKSU_CONTEXT_ERROR_WRONGPASS, "GKSU_CONTEXT_ERROR_WRONGPASS", "wrongpass" },
      { GKSU_CONTEXT_ERROR_CHILDFAILED, "GKSU_CONTEXT_ERROR_CHILDFAILED", "childfailed" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("GksuContextError", values);
  }
  return etype;
}

static gchar *
get_process_name (pid_t pid)
{
#ifdef linux
  FILE *f;
  char *cmdline = malloc (sizeof(gchar*)*256);
  char *cmdfile = g_strdup_printf("/proc/%i/status",pid);

  cmdline[255] = '\0';

  f = fopen(cmdfile, "r");
  g_free (cmdfile);
  if(f != NULL)
    {
      fgets(cmdline, 255, f);
      cmdline[255] = '\0';
      fclose(f);
    }
  return cmdline;
#else
  return NULL;
#endif
}

/**
 * prepare_xauth:
 *
 * Sets up the variables with values for the $DISPLAY
 * environment variable and xauth-related stuff. Also
 * creates a temporary directory to hold a .Xauthority
 *
 * Returns: 0 if it suceeds.
 */
static int
prepare_xauth (GksuContext *context)
{
  gchar *xauth_bin = NULL;

  FILE *xauth_output;
  gchar xauth[256] = {0};

  gchar *tmp = NULL;
  gchar **tmpv = NULL;
  gchar *display = NULL;

  /* avoid problems with "network" DISPLAY's */
  display = g_strdup (getenv ("DISPLAY"));
  tmpv = g_strsplit (display, ":", 3);
  g_free (display);

  context->display = g_strdup_printf (":%s", tmpv[1]);

  g_strfreev (tmpv);

  /* find out where the xauth binary is located */
  if (g_file_test ("/usr/bin/xauth", G_FILE_TEST_IS_EXECUTABLE))
    xauth_bin = "/usr/bin/xauth";
  else if (g_file_test ("/usr/X11R6/bin/xauth", G_FILE_TEST_IS_EXECUTABLE))
    xauth_bin = "/usr/X11R6/bin/xauth";
  else
    {
      fprintf (stderr,
	       _("Failed to obtain xauth key: xauth binary not found "
		 "at usual locations"));

      return 1;
    }

  /* get the authorization token */
  tmp = g_strdup_printf ("%s list %s | "
			 "head -1 | awk '{ print $3 }'",
			 xauth_bin,
			 context->display);
  if ((xauth_output = popen (tmp, "r")) == NULL)
    {
      fprintf (stderr,
	       _("Failed to obtain xauth key: %s"),
	       strerror(errno));
      return 1;
    }
  fread (xauth, sizeof(char), 256, xauth_output);
  xauth[strlen(xauth) - 1] = '\0';
  pclose (xauth_output);
  g_free (tmp);

  context->xauth = g_strdup (xauth);

  if (context->debug)
    {
      fprintf(stderr,
	      "xauth: -%s-\n"
	      "display: -%s-\n",
	      context->xauth, context->display);
    }

  return 0;
}

/* Write all of buf, even if write(2) is interrupted. */
static ssize_t
full_write (int d, const char *buf, size_t nbytes)
{
  ssize_t r, w = 0;

  /* Loop until nbytes of buf have been written. */
  while (w < nbytes) {
    /* Keep trying until write succeeds without interruption. */
    do {
      r = write(d, buf + w, nbytes - w);
    } while (r < 0 && errno == EINTR);

    if (r < 0) {
      return -1;
    }

    w += r;
  }

  return w;
}

static gboolean
copy (const char *fn, const char *dir)
{
  int in, out;
  int r;
  char *newfn;
  char buf[BUFSIZ] = "";

  newfn = g_strdup_printf("%s/.Xauthority", dir);

  out = open(newfn, O_WRONLY | O_CREAT | O_EXCL, 0600);
  if (out == -1)
    {
      if (errno == EEXIST)
	fprintf (stderr,
		 _("Impossible to create the .Xauthority file: a file "
		   "already exists. This might be a security issue; "
		   "please investigate or warn your system administrator."));
      else
	fprintf (stderr,
		   _("Error copying '%s' to '%s': %s"),
		   fn, dir, strerror(errno));

      return TRUE;
    }

  in = open(fn, O_RDONLY);
  if (in == -1)
    {
      fprintf (stderr,
	       _("Error copying '%s' to '%s': %s"),
	       fn, dir, strerror(errno));
      return TRUE;
    }

  while ((r = read(in, buf, BUFSIZ)) > 0)
    {
      if (full_write(out, buf, r) == -1)
	{
	  fprintf (stderr,
		   _("Error copying '%s' to '%s': %s"),
		   fn, dir, strerror(errno));
	  return TRUE;
	}
    }

  if (r == -1)
    {
      fprintf (stderr,
	       _("Error copying '%s' to '%s': %s"),
	       fn, dir, strerror(errno));
      return TRUE;
    }

  return FALSE;
}

static gboolean
sudo_prepare_xauth (GksuContext *context)
{
  gchar template[] = "/tmp/" PACKAGE "-XXXXXX";
  gboolean error_copying = FALSE;
  gchar *xauth = NULL;

  context->dir = g_strdup (mkdtemp(template));
  if (!context->dir)
    {
      fprintf (stderr, strerror(errno));
      return 1;
    }

  xauth = g_strdup(g_getenv ("XAUTHORITY"));
  if (xauth == NULL)
    xauth = g_strdup_printf ("%s/.Xauthority", g_get_home_dir());

  error_copying = copy (xauth, context->dir);
  g_free (xauth);

  if (error_copying)
    return 1;

  return 0;
}

static void
sudo_reset_xauth (GksuContext *context, gchar *xauth,
		  gchar *xauth_env)
{
  /* reset the env var as it was before or clean it  */
  if (xauth_env)
    setenv ("XAUTHORITY", xauth_env, TRUE);
  else
    unsetenv ("XAUTHORITY");

  if (context->debug)
    fprintf (stderr, "xauth: %s\nxauth_env: %s\ndir: %s\n",
	     xauth, xauth_env, context->dir);

  unlink (xauth);
  rmdir (context->dir);

  g_free (xauth);
}

gboolean
try_su_run (GksuContext *context)
{
  int fdpty;
  pid_t pid;
  int i = 0;

  gboolean need_pass = TRUE;

  pid = forkpty (&fdpty, NULL, NULL, NULL);
  if (pid == 0)
    {
      gchar **cmd = g_malloc (sizeof(gchar*)*7);

      cmd[i] = g_strdup ("/bin/su"); i++;
      if (context->login_shell)
	{
	  cmd[i] = g_strdup ("-"); i++;
	}
      cmd[i] = g_strdup (context->user); i++;
      if (context->keep_env)
	{
	  cmd[i] = g_strdup ("-p"); i++;
	}
      cmd[i] = g_strdup ("-c"); i++;

      /* needs to get X authorization prior to running the program */
      cmd[i] = g_strdup_printf ("echo > /dev/null"); i++;

      cmd[i] = NULL;

      /* executes the command */
      if (execv (cmd[0], cmd) == -1)
	{
	  fprintf (stderr,
		   _("Unable to run /bin/su: %s"),
		   strerror(errno));

	  for (i = 0 ; cmd[i] != NULL ; i++)
	    g_free (cmd[i]);
	  g_free(cmd);

	  return TRUE;
	}
    }
  else if (pid == -1)
    {
      return TRUE;
    }
  else
    {
      fd_set rfds;

      struct timeval tv;
      struct termios tio;

      gchar buf[256] = {0};
      gint status;

      FD_ZERO(&rfds);
      FD_SET(fdpty, &rfds);

      tv.tv_sec = 0;
      tv.tv_usec = 100;

      if (select (fdpty+1, &rfds, NULL, NULL, &tv) < 0)
	return TRUE;

      read (fdpty, buf, 255);
      if (context->debug)
	fprintf (stderr, "try_su_run: buf: -%s-\n", buf);

      usleep (100);

      tcgetattr (fdpty, &tio);

      if ((tio.c_lflag & ECHO))
	need_pass = FALSE;

      kill (pid, SIGKILL);
      waitpid (pid, &status, 0);
    }

  return need_pass;
}

/**
 * gksu_context_try_need_password
 * @context: a #GksuContext
 *
 * Checks if we need to ask for a password or if we have ways of
 * getting the password for ourselves or we simply don't need it.
 * Since 1.3.0.
 *
 * Returns: TRUE if requesting a password is needed, FALSE otherwise.
 *
 */
gboolean
gksu_context_try_need_password (GksuContext *context)
{
  nullify_password (context);

  if (try_su_run (context) == FALSE)
    return FALSE;

  return TRUE;
}

/**
 * gksu_context_run_full:
 * @context: a #GksuContext
 * @ask_pass: a #GksuAskPasswordFunc
 * @user_data: a #gpointer with user data to be passed to the
 * #GksuAskPasswordFunc
 * @pass_not_needed: a #GksuPasswordNotNeededFunc
 * @pnn_user_data: a #gpointer with the user data to be passed to the
 * #GksuPasswordNotNeededFunc
 * @error: a #GError object to be filled with the error code or NULL
 *
 * This could be considered one of the main functions in GKSu.
 * it is responsible for doing the 'user changing' magic calling
 * the #GksuAskPasswordFunc function to request a password if needed.
 * and the #GksuPasswordNotNeededFunc function if a password won't be
 * needed, so the application has the oportunity of warning the user
 * what it's doing.
 * Since 1.3.3.
 *
 * Returns: the child's error status, 0 if all went fine, -1 if failed
 */
gboolean
gksu_context_run_full (GksuContext *context,
		       GksuAskPasswordFunc ask_pass, gpointer user_data,
		       GksuPasswordNotNeededFunc pass_not_needed,
		       gpointer pnn_user_data, GError **error)
{
  GQuark gksu_quark;
  int i = 0;

  gchar auxcommand[] = PREFIX "/lib/" PACKAGE "/gksu-run-helper";

  int fdpty;
  pid_t pid;

  gksu_quark = g_quark_from_string (PACKAGE);

  if (!context->command)
    {
      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_NOCOMMAND,
		   _("gksu_run needs a command to be run, "
		     "none was provided."));
      return -1;
    }

  if (!g_file_test (auxcommand, G_FILE_TEST_IS_EXECUTABLE))
    {
      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_HELPER,
		   _("The gksu-run-helper command was not found or "
		     "is not executable."));
      return -1;
    }

  /*
     FIXME: need to check if we are in X and also need to
     provide GError stuff
  */
  prepare_xauth (context);

  pid = forkpty (&fdpty, NULL, NULL, NULL);
  if (pid == 0)
    {
      gchar **cmd = g_malloc (sizeof(gchar*)*7);

      cmd[i] = g_strdup ("/bin/su"); i++;
      if (context->login_shell)
	{
	  cmd[i] = g_strdup ("-"); i++;
	}
      cmd[i] = g_strdup (context->user); i++;
      if (context->keep_env)
	{
	  cmd[i] = g_strdup ("-p"); i++;
	}
      cmd[i] = g_strdup ("-c"); i++;

      /* needs to get X authorization prior to running the program */
      cmd[i] = g_strdup_printf ("%s \"%s\"", auxcommand,
				context->command); i++;

      cmd[i] = NULL;

      /* executes the command */
      if (execv (cmd[0], cmd) == -1)
	{
	  fprintf (stderr,
		   _("Unable to run /bin/su: %s"),
		   strerror(errno));
	}

      for (i = 0 ; cmd[i] != NULL ; i++)
	g_free (cmd[i]);
      g_free(cmd);
    }
  else if (pid == -1)
    {
      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_FORK,
		   _("Failed to fork new process: %s"),
		   strerror(errno));
      return -1;
    }
  else
    {
      fd_set rfds;

      struct timeval tv;

      struct passwd *pwd = NULL;
      gint target_uid = -1;
      gint my_uid = 0;

      gchar buf[256] = {0};
      gint status;

      gchar *cmdline = NULL;

      gboolean password_needed = FALSE;

      my_uid = getuid();
      pwd = getpwnam (context->user);
      if (pwd)
	target_uid = pwd->pw_uid;

      /* no need to ask for password if we're already root */
      if (my_uid != target_uid && my_uid)
	{
	  if (context->password || ask_pass)
	    {
	      struct termios tio;

	      usleep (100);
	      read (fdpty, buf, 255);
	      if (context->debug)
		fprintf (stderr, "gksu_context_run: buf: -%s-\n", buf);

	      usleep (100);

	      tcgetattr (fdpty, &tio);

	      if (!(tio.c_lflag & ECHO))
		{
		  if (!context->password)
		    {
		      if (ask_pass (context, buf, user_data, error))
			  return -1;

		      g_return_val_if_fail (context->password, -1);
		    }

		  usleep (100);

		  password_needed = TRUE;
		  write (fdpty, context->password,
			 strlen(context->password));
		}
	    }
	}

      if (!password_needed)
	pass_not_needed (context, pnn_user_data);

      if (context->debug)
	fprintf (stderr, "DEBUG (run:after-pass) buf: -%s-\n", buf);
      if (strncmp (buf, "gksu", 4) && strncmp (buf, "su", 2))
	{
	  bzero (buf, 256);
	  /* drop the \n echoed on password entry if su did request
	     a password */
	  if (password_needed)
	    read (fdpty, buf, 256);
	  read (fdpty, buf, 255);
	  if (context->debug)
	    fprintf (stderr, "DEBUG (run:post-after-pass) buf: -%s-\n", buf);
	}

      if (!strncmp (buf, "su: Authentication failure", 26))
	{
	  g_set_error (error, gksu_quark,
		       GKSU_CONTEXT_ERROR_WRONGPASS,
		       _("Wrong password."));
	  if (context->debug)
	    fprintf (stderr, "DEBUG (auth_failed) buf: -%s-\n", buf);
	}
      else if (!strncmp (buf, "gksu: waiting", 13))
	{
	  gchar *line;

	  if (context->debug)
	    fprintf (stderr, "DEBUG (gksu: waiting) buf: -%s-\n", buf);

	  line = g_strdup_printf ("gksu-run: %s\n", context->display);
	  write (fdpty, line, strlen(line));
	  g_free (line);

	  line = g_strdup_printf ("gksu-run: %s\n", context->xauth);
	  write (fdpty, line, strlen(line));
	  g_free (line);

	  tcdrain (fdpty);

	  bzero (buf, 256);
	  read (fdpty, buf, 256);
	}
      else
	{
	  gchar *emsg = g_strdup_printf (_("Failed to communicate with "
					   "gksu-run-helper.\n\n"
					   "Received:\n"
					   " %s\n"
					   "While expecting:\n"
					   " %s"), buf, "gksu: waiting");
	  g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_HELPER, emsg);
	  g_free (emsg);

	  if (context->debug)
	    fprintf (stderr, "DEBUG (failed!) buf: -%s-\n", buf);

	  return -1;
	}

      FD_ZERO(&rfds);
      FD_SET(fdpty, &rfds);

      tv.tv_sec = 0;
      tv.tv_usec = 100;

      while (!waitpid (pid, &status, WNOHANG))
	{
	  bzero (buf, 256);
	  usleep (200);

	  if (cmdline)
	    g_free (cmdline);
	  cmdline = get_process_name (pid);

	  usleep(1000);

	  if (select (fdpty+1, &rfds, NULL, NULL, &tv) < 0)
	    {
	      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_PIPEREAD,
			   _("Could not read from the pipe "
			     "with the child: %s"),
			   strerror(errno));
	      return -1;
	    }

	  if(FD_ISSET(fdpty, &rfds))
	    {
	      bzero(buf, 256);
	      read (fdpty, buf, 255);
	      fprintf (stderr, "%s", buf);
	    }
	  else
	    FD_SET(fdpty, &rfds);
	}

      while (1)
	{
	  bzero(buf, 256);
	  if(read (fdpty, buf, 255) <= 0)
	    break;
	  fprintf (stderr, "%s", buf);
	}

      if (WIFEXITED(status))
	{
	  if (WEXITSTATUS(status))
	    {
	      if(cmdline)
		{
		  /* su already exec()ed something else, don't report
		   * exit status errors in that case
		   */
		  if (!g_str_has_suffix (cmdline, "su\n"))
		    {
		      g_free (cmdline);
		      return TRUE;
		    }
		  g_free (cmdline);
		}

	      if (error == NULL)
		g_set_error (error, gksu_quark,
			     GKSU_CONTEXT_ERROR_CHILDFAILED,
			     _("su terminated with %d status"),
			     WEXITSTATUS(status));
	    }
	}
    }

  if (error)
    return TRUE;

  return FALSE;
}

/* DO NOT USE -- Deprecated */
gboolean
gksu_context_ask_and_run (GksuContext *context, GksuAskPasswordFunc ask_pass,
			  gpointer user_data, GError **error)
{
  return gksu_context_run_full (context, ask_pass, user_data, NULL, NULL, error);
}

/**
 * gksu_context_run:
 * @context: a #GksuContext
 * @error: a #GError object to be filled with the error code or NULL
 *
 * This could be considered one of the main functions in GKSu.
 * it is responsible for doing the 'user changing' magic.
 *
 * Returns: the child's error status, 0 if all went fine, -1 if failed
 */
gboolean
gksu_context_run (GksuContext *context, GError **error)
{
  return gksu_context_run_full (context, NULL, NULL, NULL, NULL, error);
}

static gboolean
ack_killed_child (gpointer pid)
{
      waitpid (GPOINTER_TO_INT(pid), NULL, 0);
      return FALSE;
}

static gboolean
try_sudo_validation (GksuContext *context)
{
  char **cmd;
  int argcount = 4;
  char buffer[256] = {0};

  pid_t pid;
  size_t r;
  FILE *infile, *outfile;
  int parent_pipe[2];	/* For talking to the parent */
  int child_pipe[2];	/* For talking to the child */

  gboolean need_pass = TRUE;

  if ((pipe(parent_pipe)) == -1)
    return TRUE;

  if ((pipe(child_pipe)) == -1)
    return TRUE;

  cmd = g_new (gchar *, argcount + 1);

  argcount = 0;

  /* sudo binary */
  cmd[argcount] = g_strdup("/usr/bin/sudo");
  argcount++;

  /* Make sudo read from STDIN */
  cmd[argcount] = g_strdup("-S");
  argcount++;

  /* Make sudo use next arg as prompt */
  cmd[argcount] = g_strdup("-p");
  argcount++;

  /* prompt */
  cmd[argcount] = g_strdup("GNOME_SUDO_PASS");
  argcount++;

  /* just try to validate the password */
  cmd[argcount] = g_strdup("-v");
  argcount++;

  cmd[argcount] = NULL;

  pid = fork();
  if (pid == -1)
    return TRUE;
  else if (pid == 0)
    {
      // Child
      close(child_pipe[1]);
      dup2(child_pipe[0], STDIN_FILENO);
      dup2(parent_pipe[1], STDERR_FILENO);

      execv(cmd[0], cmd);

      return TRUE;
    }
  else
    {
      // Parent
      close(parent_pipe[1]);

      infile = fdopen(parent_pipe[0], "r");
      if (!infile)
	return TRUE;

      outfile = fdopen(child_pipe[1], "w");
      if (!outfile)
	return TRUE;

      /*
	we are expecting to receive a GNOME_SUDO_PASS
	if we don't there are two possibilities: an error
	or a password is not needed
      */
      r = read(parent_pipe[0], buffer, 15);
      buffer[16] = '\0';

      if (context->debug)
	fprintf (stderr, "buffer: -%s-\n", buffer);

      if (strncmp(buffer, "GNOME_SUDO_PASS", 15) == 0)
	{
	  if (context->debug)
	    fprintf (stderr, "We'll need a password...\n");

	  fprintf (outfile, "\n");
	}
      else
	{
	  if (context->debug)
	    fprintf (stderr, "We won't need a password, it seems!\n");
	  need_pass = FALSE;
	}

      /* politely request that sudo GET OUT OF THE WAY */
      kill (pid, SIGKILL);

      g_timeout_add (3000, ack_killed_child, GINT_TO_POINTER(pid));
    }

  return need_pass;
}

/**
 * gksu_context_sudo_try_need_password
 * @context: a #GksuContext
 *
 * Checks if we need to ask for a password or if we have ways of
 * getting the password for ourselves or we simply don't need it.
 * Since 1.3.1.
 *
 * Returns: TRUE if requesting a password is needed, FALSE otherwise.
 *
 */
gboolean
gksu_context_sudo_try_need_password (GksuContext *context)
{
  if (!try_sudo_validation (context))
    return FALSE;

  return TRUE;
}

/**
 * gksu_context_sudo_run_full:
 * @context: a #GksuContext
 * @ask_pass: a #GksuAskPasswordFunc
 * @user_data: a #gpointer with user data to be passed to the
 * #GksuAskPasswordFunc
 * @pass_not_needed: a #GksuPasswordNotNeededFunc
 * @pnn_user_data: a #gpointer with the user data to be passed to the
 * #GksuPasswordNotNeededFunc
 * @error: a #GError object to be filled with the error code or NULL
 *
 * This could be considered one of the main functions in GKSu.
 * it is responsible for doing the 'user changing' magic calling
 * the #GksuAskPasswordFunc function to request a password if needed.
 * and the #GksuPasswordNotNeededFunc function if a password won't be
 * needed, so the application has the oportunity of warning the user
 * what it's doing.
 * Since 1.3.3.
 *
 * Returns: the child's error status, 0 if all went fine, -1 if failed
 */
gboolean
gksu_context_sudo_run_full (GksuContext *context,
			    GksuAskPasswordFunc ask_pass, gpointer user_data,
			    GksuPasswordNotNeededFunc pass_not_needed,
			    gpointer pnn_user_data, GError **error)
{
  char **cmd;
  char buffer[256] = {0};
  int argcount = 8;
  int i, j;

  GQuark gksu_quark;

  gchar *xauth = NULL,
    *xauth_env = NULL;

  pid_t pid;
  int status;
  size_t r;
  FILE *infile, *outfile;
  int parent_pipe[2];	/* For talking to the parent */
  int child_pipe[2];	/* For talking to the child */

  gksu_quark = g_quark_from_string (PACKAGE);

  if (!context->command)
    {
      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_NOCOMMAND,
		   _("gksu_sudo_run needs a command to be run, "
		     "none was provided."));
      return -1;
    }

  /*
     FIXME: need to check if we are in X
  */
  if (sudo_prepare_xauth (context) == 1)
    {
      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_XAUTH,
		   _("Unable to copy the user's Xauthorization file."));
      return -1;
    }

  /* sets XAUTHORITY */
  xauth = g_strdup_printf ("%s/.Xauthority", context->dir);
  xauth_env = getenv ("XAUTHORITY");
  setenv ("XAUTHORITY", xauth, TRUE);

  if (context->debug)
    fprintf (stderr, "xauth: %s\n", xauth);

  cmd = g_new (gchar *, argcount + 1);

  argcount = 0;

  /* sudo binary */
  cmd[argcount] = g_strdup("/usr/bin/sudo");
  argcount++;

  if (!context->keep_env)
    {
      /* Make sudo set $HOME */
      cmd[argcount] = g_strdup("-H");
      argcount++;
    }

  /* Make sudo read from stdin */
  cmd[argcount] = g_strdup("-S");
  argcount++;

  /* Make sudo use next arg as prompt */
  cmd[argcount] = g_strdup("-p");
  argcount++;

  /* prompt */
  cmd[argcount] = g_strdup("GNOME_SUDO_PASS");
  argcount++;

  /* Make sudo use the selected user */
  cmd[argcount] = g_strdup("-u");
  argcount++;

  /* user */
  cmd[argcount] = g_strdup(context->user);
  argcount++;

  /* sudo does not understand this if we do not use -H
     weird.
  */
  if (!context->keep_env)
    {
      /* Make sudo stop processing options */
      cmd[argcount] = g_strdup("--");
      argcount++;
    }

  {
    gchar *tmp_arg = g_malloc (sizeof(gchar)*1);
    gboolean inside_quotes = FALSE;

    tmp_arg[0] = '\0';

    for (i = j = 0; ; i++)
      {
	if ((context->command[i] == '\'') && (context->command[i-1] != '\\'))
	  {
	    i = i + 1;
	    inside_quotes = !inside_quotes;
	  }

	if ((context->command[i] == ' ' && inside_quotes == FALSE)
	    || context->command[i] == '\0')
	  {
	    tmp_arg = g_realloc (tmp_arg, sizeof(gchar)*(j+1));
	    tmp_arg[j] = '\0';
	    cmd = g_realloc (cmd, sizeof(gchar*) * (argcount + 1));
	    cmd[argcount] = g_strdup (tmp_arg);

	    g_free (tmp_arg);

	    argcount = argcount + 1;
	    j = 0;

	    if (context->command[i] == '\0')
	      break;

	    tmp_arg = g_malloc (sizeof(gchar)*1);
	    tmp_arg[0] = '\0';
	  }
	else
	  {
	    if (context->command[i] == '\\')
	      i = i + 1;
	    tmp_arg = g_realloc (tmp_arg, sizeof(gchar)*(j+1));
	    tmp_arg[j] = context->command[i];
	    j = j + 1;
	  }
      }
  }
  cmd = g_realloc (cmd, sizeof(gchar*) * (argcount + 1));
  cmd[argcount] = NULL;

  if (context->debug)
    {
      for (i = 0; cmd[i] != NULL; i++)
	fprintf (stderr, "cmd[%d]: %s\n", i, cmd[i]);
    }

  if ((pipe(parent_pipe)) == -1)
    {
      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_PIPE,
		   _("Error creating pipe: %s"),
		   strerror(errno));
      sudo_reset_xauth (context, xauth, xauth_env);
      return -1;
    }

  if ((pipe(child_pipe)) == -1)
    {
      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_PIPE,
		   _("Error creating pipe: %s"),
		   strerror(errno));
      sudo_reset_xauth (context, xauth, xauth_env);
      return -1;
    }

  pid = fork();
  if (pid == -1)
    {
      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_FORK,
		   _("Failed to fork new process: %s"),
		   strerror(errno));
      sudo_reset_xauth (context, xauth, xauth_env);
      return -1;
    }
  else if (pid == 0)
    {
      // Child
      setsid();   // make us session leader
      close(child_pipe[1]);
      dup2(child_pipe[0], STDIN_FILENO);
      dup2(parent_pipe[1], STDERR_FILENO);

      execv(cmd[0], cmd);

      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_EXEC,
		   _("Failed to exec new process: %s"),
		   strerror(errno));
      sudo_reset_xauth (context, xauth, xauth_env);
      return -1;
    }
  else
    {
      gchar *cmdline = NULL;

      // Parent
      close(parent_pipe[1]);

      infile = fdopen(parent_pipe[0], "r");
      if (!infile)
	{
	  g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_PIPE,
		       _("Error opening pipe: %s"),
		       strerror(errno));
	  sudo_reset_xauth (context, xauth, xauth_env);
	  return -1;
	}

      outfile = fdopen(child_pipe[1], "w");
      if (!outfile)
	{
	  g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_PIPE,
		       _("Error opening pipe: %s"),
		       strerror(errno));
	  sudo_reset_xauth (context, xauth, xauth_env);
	  return -1;
	}

      /*
	we are expecting to receive a GNOME_SUDO_PASS
	if we don't there are two possibilities: an error
	or a password is not needed
      */
      r = read(parent_pipe[0], buffer, 16);

      if (context->debug)
	fprintf (stderr, "buffer: -%s-\n", buffer);

      if (strncmp(buffer, "GNOME_SUDO_PASS", 15) == 0)
	{
	  if (context->debug)
	    fprintf (stderr, "Yeah, we're in...\n");

	  if ((context->password == NULL || (!strcmp (context->password, "")))
	      && ask_pass == NULL)
	    {
	      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_NOPASSWORD,
			   _("No password was supplied and sudo needs it."));
	      fprintf (outfile, "\n");
	      sudo_reset_xauth (context, xauth, xauth_env);
	      return -1;
	    }

	  if (!context->password)
	    {
	      if (ask_pass (context, _("Password: "), user_data, error))
		return -1;

	      g_return_val_if_fail (context->password, -1);
	    }

	  usleep (100);

	  fprintf (outfile, "%s", context->password);
	  fclose (outfile);

	  /* ignore the newline that comes right after sudo receives
	     the password
	  */
	  fgets (buffer, 255, infile);
	  fgets (buffer, 255, infile);
	}
      else
	{
	  if (context->debug)
	    fprintf (stderr, "No password prompt found; we'll assume we don't need a password.\n");

	  if (pass_not_needed)
	    pass_not_needed (context, pnn_user_data);

	  fprintf (stderr, "%s", buffer);
	}

      if (!strcmp (buffer, "Sorry, try again.\n"))
	g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_WRONGPASS,
		     _("Wrong password."));
      else if (!strncmp (buffer, "Sorry, user ", 12))
	g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_NOT_ALLOWED,
		     _("The underlying authorization mechanism (sudo) "
		       "does not allow you to run this program. Contact "
		       "the system administrator."));
      else
	{
	  gchar *haystack = buffer;
	  gchar *needle;

	  needle = g_strstr_len (haystack, strlen (haystack), " ");
	  if (needle && (needle + 1))
	    {
	      needle += 1;
	      if (!strncmp (needle, "is not in", 9))
		g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_NOT_ALLOWED,
			     _("The underlying authorization mechanism (sudo) "
			       "does not allow you to run this program. Contact "
			       "the system administrator."));
	    }
	}

      while (!waitpid (pid, &status, WNOHANG))
	{
	  if (cmdline)
	    g_free (cmdline);
	  cmdline = get_process_name (pid);

	  usleep(1000);

	  bzero(buffer, 256);
	  if(!fgets (buffer, 255, infile))
	    continue;
	  fprintf (stderr, "%s", buffer);
	}

      /* make sure we did read everything */
      while (1)
	{
	  bzero(buffer, 256);
	  if(!fread (buffer, sizeof(gchar), 255, infile))
	    break;
	  fprintf (stderr, "%s", buffer);
	  fflush (stderr);
	}

      sudo_reset_xauth (context, xauth, xauth_env);

      if (WIFEXITED(status))
	{
	  if (WEXITSTATUS(status))
	    {
	      if(cmdline)
		{
		  /* sudo already exec()ed something else, don't report
		   * exit status errors in that case
		   */
		  if (!g_str_has_suffix (cmdline, "sudo\n"))
		    {
		      g_free (cmdline);
		      return TRUE;
		    }
		  g_free (cmdline);
		}

	      if (error == NULL)
		g_set_error (error, gksu_quark,
			     GKSU_CONTEXT_ERROR_CHILDFAILED,
			     _("sudo terminated with %d status"),
			     WEXITSTATUS(status));
	    }
	}
    }

  /* if error is set we have found an error condition */
  if (error)
    return TRUE;

  return FALSE;
}

/* DO NOT USE -- Deprecated */
gboolean
gksu_context_sudo_ask_and_run (GksuContext *context, GksuAskPasswordFunc ask_pass,
			       gpointer user_data, GError **error)
{
  return gksu_context_sudo_run_full (context, ask_pass, user_data, NULL, NULL, error);
}

/**
 * gksu_context_sudo_run:
 * @context: a #GksuContext
 * @error: a #GError object to be filled with the error code or NULL
 *
 * This could be considered one of the main functions in GKSu.
 * it is responsible for doing the 'user changing' magic by
 * calling gksu_ask_password() if it needs the user's password
 * it behaves like sudo.
 *
 * Returns: the child's error status, 0 if all went fine, -1 if failed
 */
gboolean
gksu_context_sudo_run (GksuContext *context, GError **error)
{
  return gksu_context_sudo_run_full (context, NULL, NULL, NULL, NULL, error);
}
